import os.path
from io import StringIO

import django
from django.apps import apps
from django.conf import settings
from django.core.management import call_command
from django.db.utils import OperationalError
from qgis.core import (
    Qgis,
    QgsCoordinateReferenceSystem,
    QgsDataSourceUri,
    QgsMessageLog,
    QgsProject,
    QgsProviderMetadata,
    QgsProviderRegistry,
    QgsVectorLayer,
)
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtWidgets import QAction, QMessageBox

from . import datamodels_registry, prepare_django, register_datamodel
from .contrib.demo_models import config
from .provider import Provider
from .utils import find_geom_field, find_pk_field

QgsMessageLog.logMessage("loading qdmtk file", "QDMTK")


class Plugin:
    """QGIS Plugin Implementation."""

    def __init__(self, iface):
        self.iface = iface
        self.plugin_dir = os.path.dirname(__file__)

        # Load the demo datamodel
        register_datamodel(config.PROJECTNAME_A, config.APPS_A, config.DATABASE_A)
        register_datamodel(config.PROJECTNAME_B, config.APPS_B, config.DATABASE_B)

        iface.initializationCompleted.connect(prepare_django)

    def initGui(self):
        """Create the menu entries and toolbar icons inside the QGIS GUI."""

        # Register our provider
        metadata = QgsProviderMetadata(
            Provider.providerKey(), Provider.description(), Provider.createProvider
        )
        QgsProviderRegistry.instance().registerProvider(metadata)

        # Add toolbar
        self.toolbar = self.iface.addToolBar("Datamodel")

        self.load_layers_dj_action = QAction(
            QIcon(os.path.join(self.plugin_dir, "icon_dj.svg")),
            "Load datamodel layers (using Django provider)",
            self.toolbar,
        )
        self.load_layers_dj_action.triggered.connect(self.load_layers_dj)
        self.toolbar.addAction(self.load_layers_dj_action)

        self.load_layers_pg_action = QAction(
            QIcon(os.path.join(self.plugin_dir, "icon_pg.svg")),
            "Load datamodel layers (using Postgres provider)",
            self.toolbar,
        )
        self.load_layers_pg_action.triggered.connect(self.load_layers_pg)
        self.toolbar.addAction(self.load_layers_pg_action)

        self.migrate_action = QAction("Migrate", self.toolbar)
        self.migrate_action.triggered.connect(self.migrate)
        self.toolbar.addAction(self.migrate_action)

        self.showmigrations_action = QAction("Show migrations", self.toolbar)
        self.showmigrations_action.triggered.connect(self.showmigrations)
        self.toolbar.addAction(self.showmigrations_action)

        # Uncomment this to load the demo datamodel
        # from . import register_datamodel
        # from .qdmtkdemo import config
        # register_datamodel(config.DATAMODEL_NAME, config.INSTALLED_APPS, config.DATABASE)

    def unload(self):
        self.iface.mainWindow().removeToolBar(self.toolbar)
        # seems this does not exist ? can we still autoreload
        # QgsProviderRegistry.instance().unregisterProvider(Provider.providerKey())

    def migrate(self):
        apps_names = {app.name: app.label for app in apps.get_app_configs()}

        out = StringIO()
        for datamodel_key, datamodel_opts in datamodels_registry.items():
            apps_to_migrate = [
                apps_names[app_name] for app_name in datamodel_opts["apps"]
            ]
            out.write(f"--- {datamodel_key} ---\n")
            for app_to_migrate in apps_to_migrate:
                # app_labels = datamodels_registry[datamodel_key]["apps"]
                QgsMessageLog.logMessage(
                    f"Will migrate {app_to_migrate} to {datamodel_key}", "QDMTK"
                )
                try:
                    call_command(
                        "migrate",
                        app_to_migrate,
                        "--database",
                        datamodel_key,
                        stdout=out,
                    )
                except OperationalError as e:
                    out.write(f"Could not connect ({e})\n")
                out.flush()
        QMessageBox.information(
            self.iface.mainWindow(), "Migration results", out.getvalue()
        )

    def showmigrations(self):
        apps_names = {app.name: app.label for app in apps.get_app_configs()}
        out = StringIO()
        for datamodel_key, datamodel_opts in datamodels_registry.items():
            apps_to_migrate = [
                apps_names[app_name] for app_name in datamodel_opts["apps"]
            ]
            out.write(f"--- {datamodel_key} ---\n")
            try:
                call_command(
                    "showmigrations",
                    *apps_to_migrate,
                    "--database",
                    datamodel_key,
                    stdout=out,
                )
            except OperationalError as e:
                out.write(f"Could not connect ({e})\n")
            out.flush()
        QMessageBox.information(
            self.iface.mainWindow(), "Current migrations state", out.getvalue()
        )

    def load_layers_dj(self, checked):
        for model in django.apps.apps.get_models():
            if not getattr(model, "qdmtk_addlayer", False):
                continue
            layer = QgsVectorLayer(
                model.__name__, model.__name__, Provider.providerKey()
            )
            QgsProject.instance().addMapLayer(layer)

        self.iface.messageBar().pushMessage(
            "Success", "Loaded Datamodel", level=Qgis.Info
        )

    def load_layers_pg(self, checked):
        for model in django.apps.apps.get_models():
            if not getattr(model, "qdmtk_addlayer", False):
                continue
            db = settings.DATABASES["default"]
            uri = QgsDataSourceUri()
            uri.setConnection(
                db["HOST"], db["PORT"], db["NAME"], db["USER"], db["PASSWORD"]
            )
            query = f"({model.objects.all().query})"
            geom_field = find_geom_field(model)
            geom_field.srid if geom_field else None
            key_field = getattr(find_pk_field(model), "name", None)
            uri.setDataSource(
                None,
                query,
                geom_field.name if geom_field else None,
                aKeyColumn=key_field,
            )
            uri.setUseEstimatedMetadata(True)
            if geom_field:
                uri.setSrid(str(geom_field.srid))
            layer = QgsVectorLayer(uri.uri(), model.__name__, "postgres")
            if geom_field:
                crs = QgsCoordinateReferenceSystem()
                crs.createFromString(f"POSTGIS:{geom_field.srid}")
                layer.setCrs(crs)

            # Loading a postgres current doesn't work because of `ERROR: function st_srid(bytea) is not unique`
            # when loaded in QGIS. Seems to happen on layer loading when SRID is unknown, due to some ambigous
            # type casts made by geodjango.
            # See https://github.com/qgis/QGIS/blob/082aa7bbcb847b9ed507e808203bb23c979c4c45/src/providers/postgres/qgspostgresconn.cpp#L1966-L1972
            # Suprisingly, setting the CRS as done above is not enough and this code still runs.
            # We work-around this by monkey-patching the cast, but that breaks django' provider :-/
            # TODO : find a fix to remove this monkey patch (either upstream in geodjango to have a better cast,
            # or in QGIS to avoid this query being run)
            from django.contrib.gis.db.backends.postgis.operations import (
                PostGISOperations,
            )

            PostGISOperations.select = "%s::geometry"

            QgsProject.instance().addMapLayer(layer)

        self.iface.messageBar().pushMessage(
            "Success", "Loaded Datamodel", level=Qgis.Info
        )



"""
Adapted from example in https://github.com/qgis/QGIS/blob/master/tests/src/python/provider_python.py

"""

import django
import django.conf
from django.contrib.gis.db import models
from django.contrib.gis.db.models import Extent
from django.contrib.gis.geos import GEOSGeometry
from qgis.core import (
    NULL,
    QgsAbstractFeatureIterator,
    QgsAbstractFeatureSource,
    QgsCoordinateReferenceSystem,
    QgsDataProvider,
    QgsFeatureIterator,
    QgsFeatureRequest,
    QgsField,
    QgsFields,
    QgsGeometry,
    QgsMessageLog,
    QgsRectangle,
    QgsVectorDataProvider,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QVariant

from .utils import find_geom_field, string_to_fid


class FeatureIterator(QgsAbstractFeatureIterator):
    def __init__(self, source, request):
        super().__init__(request)
        self.request = request
        self.source = source

        self.rewind()

    def fetchFeature(self, f):
        """fetch next feature, return true on success"""
        try:
            row = next(self.iterator)

            # Geometry
            geom = QgsGeometry()
            row_geom = getattr(row, self.source.provider._geom_field.name)
            if row_geom is not None:
                geom.fromWkb(row_geom.wkb.tobytes())
            f.setGeometry(geom)
            # self.geometryToDestinationCrs(f, self._transform)

            # Fields
            f.setFields(self.source.provider.fields())
            for attr in self.source.provider._attrs():
                f.setAttribute(attr.name, getattr(row, attr.name))

            f.setValid(True)
            id = string_to_fid(row.id) if isinstance(row.id, str) else row.id
            if id is not None:
                f.setId(id)
            return True
        except StopIteration:
            f.setValid(False)
            return False

    def __iter__(self):
        """Returns self as an iterator object"""
        return self

    def __next__(self):
        """Returns the next value till current is lower than high"""
        return next(self.iterator)

    def rewind(self):
        """reset the iterator to the starting position"""

        query = self.source.provider.model.objects

        filter_rect = self.request.filterRect()
        if not filter_rect.isNull():
            # TODO : probably need to transform rect CRS if needed like this
            # transform = QgsCoordinateTransform()
            # if self.request.destinationCrs().isValid() and self.request.destinationCrs() != self.source.provider.crs():
            #     transform = QgsCoordinateTransform(self.source.provider.crs(), self.request.destinationCrs(), self.request.transformContext())
            lookup = f"{self.source.provider._geom_field.name}__bboverlaps"
            query = query.filter(**{lookup: GEOSGeometry(filter_rect.asWktPolygon())})

        if self.request.filterType() == QgsFeatureRequest.FilterType.FilterFid:
            query = query.filter(id=self.request.filterFid())

        # TODO : implement rest of filter, such as order_by, etc. (and expression ?)

        self.iterator = iter(query.all())
        return True

    def close(self):
        """end of iterating: free the resources / lock"""
        self.iterator = None
        return True


class FeatureSource(QgsAbstractFeatureSource):
    def __init__(self, provider):
        super().__init__()
        self.provider = provider

    def getFeatures(self, request):
        return QgsFeatureIterator(FeatureIterator(self, request))


class Provider(QgsVectorDataProvider):
    @classmethod
    def providerKey(cls):
        """Returns the memory provider key"""
        return "qdmtk_provider"

    @classmethod
    def description(cls):
        """Returns the memory provider description"""
        return "QDMTK Provider"

    @classmethod
    def createProvider(cls, uri, providerOptions, flags=QgsDataProvider.ReadFlags()):
        return Provider(uri, providerOptions, flags)

    # Implementation of functions from QgsVectorDataProvider

    def __init__(
        self,
        uri="",
        providerOptions=QgsDataProvider.ProviderOptions(),
        flags=QgsDataProvider.ReadFlags(),
    ):
        super().__init__(uri)

        self.uri = uri

        for model in django.apps.apps.get_models():
            if model.__name__ == self.uri:
                self.model = model
                break
        else:
            known_models = [model.__name__ for model in django.apps.apps.get_models()]
            raise Exception(
                f"Could not find model {self.uri}. Known models are {known_models}"
            )

        # Find the first geometry field
        self._geom_field = find_geom_field(self.model)

        self._extent = None

    def featureSource(self):
        return FeatureSource(self)

    def dataSourceUri(self, expandAuthConfig=True):
        return self.uri

    def storageType(self):
        return self.providerKey()

    def getFeatures(self, request=QgsFeatureRequest()):
        return self.featureSource().getFeatures(request)

    def uniqueValues(self, fieldIndex, limit=1):
        field = self._attrs_map()[fieldIndex]
        return self.model.objects.values_list(field, flat=True).distinct()

    def wkbType(self):
        if self._geom_field is None:
            return QgsWkbTypes.NoGeometry
        return QgsWkbTypes.parseType(self._geom_field.geom_type)

    def featureCount(self):
        return self.model.objects.count()

    def _attrs(self):
        """
        Yields SQLAlchemy's attributes (non-geom)
        """
        for field in self.model._meta.get_fields():
            if isinstance(field, (models.OneToOneField, models.ManyToOneRel)):
                # Skip non-field attributes
                continue
            if field is self._geom_field:
                # Skip the geometry field, which is not an attribute
                continue
            yield field

    def _attrs_map(self) -> "dict[int, str]":
        """
        Returns a dict that maps field idx to field name
        """
        return {i: attr.name for i, attr in enumerate(self._attrs())}

    def fields(self):
        fields = QgsFields()
        for attr in self._attrs():
            if isinstance(attr, models.ForeignKey):
                field = attr.target_field
            else:
                field = attr

            if isinstance(field, (models.TextField, models.CharField)):
                type_ = QVariant.String
            elif isinstance(field, models.IntegerField):
                type_ = QVariant.Int
            elif isinstance(field, (models.FloatField, models.DecimalField)):
                type_ = QVariant.Double
            elif isinstance(field, models.DateField):
                type_ = QVariant.Date
            elif isinstance(field, models.BooleanField):
                type_ = QVariant.Bool
            else:
                QgsMessageLog.logMessage(
                    f"Field type not configured : {field.__class__.__name__} ({attr.name})",
                )
                type_ = QVariant.Invalid
            fields.append(QgsField(attr.name, type_))
        return fields

    def addFeatures(self, flist, flags=None):
        for f in flist:
            instance = self.model()
            setattr(instance, self._geom_field.name, GEOSGeometry(f.geometry().asWkt()))
            for attr in self._attrs():
                value = f.attribute(attr.name)
                setattr(instance, attr.name, value if value != NULL else None)
            instance.save()
        return True, flist

    def deleteFeatures(self, ids):
        for instance in self.model.objects.filter(id__in=ids):
            instance.delete()

    def changeAttributeValues(self, attr_map):
        attrs_map = self._attrs_map()
        for fid, attrs in attr_map.items():
            instance = self.model.objects.get(id=fid)
            for k, value in attrs.items():
                setattr(instance, attrs_map[k], value if value != NULL else None)
            instance.save()
        return True

    def changeGeometryValues(self, geometry_map):
        for fid, geometry in geometry_map.items():
            instance = self.model.objects.get(id=fid)
            setattr(instance, self._geom_field.name, GEOSGeometry(geometry.asWkt()))
            instance.save()
        return True

    def allFeatureIds(self):
        return self.model.objects.values_list("id", flat=True)

    def subsetString(self):
        return ""

    def setSubsetString(self, subsetString):
        raise NotImplementedError()

    def supportsSubsetString(self):
        return False

    def capabilities(self):
        return (
            QgsVectorDataProvider.AddFeatures
            | QgsVectorDataProvider.DeleteFeatures
            | QgsVectorDataProvider.ChangeGeometries
            | QgsVectorDataProvider.ChangeAttributeValues
            | QgsVectorDataProvider.SelectAtId
        )

    # /* Implementation of functions from QgsDataProvider */

    def name(self):
        return self.providerKey()

    def extent(self):
        if self._extent is None:
            self.updateExtents()
        return self._extent

    def updateExtents(self):
        extents = self.model.objects.aggregate(extent=Extent(self._geom_field.name))[
            "extent"
        ]
        if extents:
            self._extent = QgsRectangle(extents[0], extents[1], extents[2], extents[3])
        else:
            self._extent = QgsRectangle()

    def isValid(self):
        return True

    def crs(self):
        crs = QgsCoordinateReferenceSystem()
        srid = self._geom_field.srid
        crs.createFromString(f"EPSG:{srid}")
        return crs

    def handlePostCloneOperations(self, source):
        pass



from django.apps import apps

from . import datamodels_registry


class QDMTKDatabaseRouter:
    """
    This routes each model to the database corresponding to the datamodel that installed this model
    """

    registry = {}

    def db_for_read(self, model, **hints):
        app = apps.get_app_config(model._meta.app_label)
        for datamodel_key in datamodels_registry:
            if app.name in datamodels_registry[datamodel_key]["apps"]:
                return datamodel_key
        raise Exception(f"No datamodel found {model} (of app {model._meta.app_label})")

    def db_for_write(self, model, **hints):
        return self.db_for_read(model, **hints)

    def allow_migrate(self, db, app_label, model_name=None, **hints):
        # NOTE : This isn't well used by Django. Even when migrations are not allowed against a database,
        # the corresponding migrations re inserted in the django_migrations tables and listed in
        # django showmigrations.
        # See https://code.djangoproject.com/ticket/23273

        # When migrate is run on a DB that doens't belong to the datamodels (for instance
        # if migrate was called without --database in which case db='default'), we skip the model
        if db not in datamodels_registry:
            return False

        # Otherwise, we only run if the model's app is listed as an installed app of the given
        # datamodel.
        app = apps.get_app_config(app_label)
        return app.name in datamodels_registry[db]["apps"]



import os

from .version import __version__  # noqa

# This will hold all registered datamodels
datamodels_registry = {}


def classFactory(iface):
    from .plugin import Plugin

    return Plugin(iface)


def register_datamodel(datamodel_key, installed_apps, db_settings):
    """
    This registers the database connection and installed apps for the given datamodel.
    These will be used to configure Django when prepare_django() is called.
    """
    from django.apps import apps

    from .exceptions import QDMTKException

    if apps.ready:
        raise QDMTKException(
            "Django was already setup. Ensure `prepare_django` is only called after all datamodels are registered."
        )

    datamodels_registry[datamodel_key] = {
        "apps": installed_apps,
        "db_settings": db_settings,
    }


def prepare_django():
    """
    Sets up Django with all registered apps and database settings
    """
    import django
    from django.apps import apps
    from django.conf import settings

    if apps.ready:
        # already done
        return

    # Allow to configure GDAL/GEOS/Spatialite libraries from env vars
    # see https://docs.djangoproject.com/en/3.2/ref/contrib/gis/install/geolibs/#geos-library-path
    GDAL_LIBRARY_PATH_ENV = os.getenv("GDAL_LIBRARY_PATH")
    GEOS_LIBRARY_PATH_ENV = os.getenv("GEOS_LIBRARY_PATH")
    SPATIALITE_LIBRARY_PATH_ENV = os.getenv("SPATIALITE_LIBRARY_PATH")
    additionnal_settings = {}
    if GDAL_LIBRARY_PATH_ENV:
        additionnal_settings["GDAL_LIBRARY_PATH"] = GDAL_LIBRARY_PATH_ENV
    if GEOS_LIBRARY_PATH_ENV:
        additionnal_settings["GEOS_LIBRARY_PATH"] = GEOS_LIBRARY_PATH_ENV
    if SPATIALITE_LIBRARY_PATH_ENV:
        additionnal_settings["SPATIALITE_LIBRARY_PATH"] = SPATIALITE_LIBRARY_PATH_ENV

    # Collect all databases
    databases = {k: v["db_settings"] for k, v in datamodels_registry.items()}
    # Contactenate apps from all datamodels
    installed_apps = []
    for datamodel in datamodels_registry.values():
        installed_apps.extend(datamodel["apps"])

    # Set a default database if needed (required by django)
    if "default" not in databases:
        databases["default"] = {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": ":memory:",
        }

    # from qgis.core import QgsMessageLog
    # QgsMessageLog.logMessage(f"Setting up Django with {databases=} {installed_apps=}", "QDMTK")

    settings.configure(
        DATABASES=databases,
        INSTALLED_APPS=installed_apps,
        DATABASE_ROUTERS=["qdmtk.router.QDMTKDatabaseRouter"],
        **additionnal_settings,
    )
    django.setup()



from django.contrib.gis.db import models


def string_to_fid(string):
    """
    Converts a string to a longlong, matching QGIS's STRING_TO_FID macro
    """
    try:
        return int(string)
    except ValueError:
        return 0


def find_geom_field(model):
    for field in model._meta.get_fields():
        if isinstance(field, models.GeometryField):
            return field
    return None


def find_pk_field(model):
    return model._meta.pk



import os
from distutils.core import setup

from setuptools import find_packages, setup

from qdmtk import __version__

requirements_path = os.path.join("qdmtk", "requirements.txt")

setup(
    name="QDMTK",
    version=__version__,
    description="QGIS Datamodel toolkit",
    author="Olivier Dalang",
    author_email="olivier@opengis.ch",
    url="https://github.com/opengisch/qdmtk",
    packages=find_packages(),
    install_requires=open(requirements_path, encoding="utf-8").read().splitlines(),
    include_package_data=True,
)
